:- use_module( library(term_type) ). % /2
:- requires( nth_arg/4 ).

%% position( Type, N, Data, Nth ).
%% position( N, Data, Nth ).
%
%  An experimental polymorphic predicate that works on Data that is one of,
%  list, atomic or compound. When atomic only position 1 is valid.
%  
% ==
%  position( 2, [1,2,3], W ).
%  position( 2, c(1,2,3), W ).
%  position( compound, 2, c(1,2,3), W ).
%  position( list, 2, c(1,2,3), W ).
%  position( list, 2, c(1,2,3), W ).
% == 
% @author  nicos angelopoulos
% @version 0.1 2014/02/09
% @version 0.2 2014/06/30 switch to term_type/2.
%
position( N, Data, Nth ) :-
	term_type( Data, Dtype ),
	% position_type( Data, Dtype ),
	position( Dtype, N, Data, Nth ).

position( number, 1, Data, Data ).
position( atom, 1, Data, Data ).
position( list, N, Data, Nth ) :-
	nth1( N, Data, Nth ).
position( compound, N, Data, Nth ) :-
	arg( N, Data, Nth ).

%% n_positions( +N, +Data, -FirstN ).
% 
% Get the first N positions from Data into FirstN.
%
%==
% ?- n_positions( 3, [a,b,c,d], First ).
% First = [a, b, c].
% ?- n_positions( 5, [a,b,c,d], First ).
% First = [a, b, c, d].
% ?- n_positions( 3, t(a,b,c,d), First ).
% First = t(a, b, c).
% ?- n_positions( 3, "abcd", First ).
% First = "abc".
% ?- n_positions( 3, 'abcd', First ).
% First = abc.
% 
%==
% @author nicos angelopoulos
% @version  0.1 2014/6/30
%
n_positions( N, DataIn, FirstN ) :-
	term_type( DataIn, TypeIn ),
	number_type_atom( TypeIn, DataIn, Type, Data ),
	n_positions_type( N, Data, Type, Elements ),
	type_term_elements_term( Type, Data, Elements, FirstNOut ),
	number_type_converts( TypeIn, FirstNOut, FirstN ).

n_positions_type( N, Data, Type, Elems ) :-
	N > 0, % N reached, or
	position_nth( Type, 1, Data, Ith, Rem, _J ),
	!, % otherwise we are at the end, before N elements were found
	M is N - 1,
	Elems = [Ith|TElems],
	n_positions_type( M, Rem, Type, TElems ).
n_positions_type( _, _Data, _Type, [] ).
	% Acc = FirstN.

%% position_nth( +N, +Data, -Nth ).
%% position_nth( +N, +Data, -Nth, -Rem, -Nxt ).
%% position_nth( +Dtype, +N, +Data, -Nth, -Rem, -Nxt ).
%
% Get Data's N position datum into Nth, 
% with Rem being what is left of data and Nxt is the
% N identifier for the next to the _right_ of Nth.
% Predicate expects that bounds are respected, else fails.
% Dtype is the datatype of Data, either _list_ or 
% _compound_ which is determined by the predicate if missing.
%
% ==
%    Data = [1,2,3,4,5],
%    position_nth( list, 2, Data, Nth, Rem, Nxt ).
%    position_nth( compound, 2, Data, Nth, Rem, Nxt ).
%    position_nth( list, 1, Data, Nth, Rem, Nxt ).
%
% ?- maplist( position_nth(3), [c(1,2,3),c(4,5,6)], Thirds, Rem ).
% Thirds = [3, 6],
% Rem = [c(1, 2), c(4, 5)].
% ==
% @author  nicos angelopoulos
% @version 0.2, 2014/02/27 changed from position_next
%
position_nth( N, Data, Nth ) :-
	term_type( Data, Dtype ),
	position_type_nth( Dtype, N, Data, Nth ).

position_nth( N, Data, Nth, Rem ) :-
	term_type( Data, Dtype ),
	position_nth( Dtype, N, Data, Nth, Rem, _Nxt ).

position_nth( N, Data, Nth, Rem, Nxt ) :-
	term_type( Data, Dtype ),
	position_nth( Dtype, N, Data, Nth, Rem, Nxt ).

position_nth( number, N, Data, Nth, Rem, N ) :-
	number_codes( Data, Codes ),
	position_nth_list( N, Codes, Nth, RemCodes, _Nxt ),
	number_codes( Rem, RemCodes ).

position_nth( atom, N, Data, Nth, Rem, N ) :-
	atom_codes( Data, Codes ),
	position_nth_list( N, Codes, Nth, RemCodes, _Nxt ),
	atom_codes( Rem, RemCodes ).
position_nth( list, N, Data, Nth, Rem, Nxt ) :-
	position_nth_list( N, Data, Nth, Rem, Nxt ).
position_nth( string, N, Data, Nth, Rem, N ) :-
	string_codes( Data,  Codes ),
	position_nth_list( N, Codes, Nth, RemCodes, _Nxt ),
	string_codes( Rem, RemCodes ).
position_nth( compound, N, Data, Nth, Rem, N ) :-
	nth_arg( N, Data, Nth, Rem ).

% special case
% position_nth_list( 1, [H|T], H, T, 2 ) :- !.
position_nth_list( N, List, Nth, Rest, N ) :-
	nth1( N, List, Nth, Rest ).

% check the modes, probably works fine for generating.
position_type_nth( list, N, Data, Nth ) :-
	nth1( N, Data, Nth ).
position_type_nth( number, 1, Datum, Datum ).
position_type_nth( atom, 1, Datum, Datum ).
position_type_nth( compound, N, Data, Nth ) :-
	arg( N, Data, Nth ).

%% positions( +Data, -Dtype, -NofPositions ).
%% positions( +Data, -NofPositions ).
%
%  Number of positions and data type for list/compound Data.
% If Data is a list NofPositions is the length. If Data is atomic
% the length is 1, and otherwise the number of positions
% is its arity. Dtype is correspondingly, _list_ and  _compound_.
% 
%==
%   positions( [1,2,3,4], P ).
%==
% 
% @author nicos angelopoulos
% @version  0.1 2014/02/09
% @tbd  allow for data()  (see my compound preds).
%
positions( Data, Positions ) :-
	positions( Data, _Dtype, Positions ).

positions( Data, Dtype, Positions ) :-
	position_type( Data, Dtype ),
	positions_type( Dtype, Data, Positions ).

positions_type( atomic, _Data, 1 ).
positions_type( list, Data, Positions ) :-
	length( Data, Positions ).
positions_type( compound, Data, Positions ) :-
	functor( Data, _Name, Positions ).

%% position_type( +Data, -Dtype ).
%
%  Dtype is the determined datatype for Data.
%  If atomic(Data) succeeds, Dtype is _atomic_.
%  If Dtype is not a variable and it unifies [_|_],
%  then Dtype unifies _list_, Otherwise, 
%  Dtype is _compound_.
%
position_type( Data, Dtype ) :-
	atomic( Data ),
	!,
	Dtype = atomic.
position_type( Data, Dtype ) :-
	\+ var( Data ),
	Data = [_|_],
	!,
	Dtype = list.
position_type( _Data, Dtype ) :-
	% test \+ var( Data ),  ??
	Dtype = compound.

%% positions_term( +Data, -Fresh ).
%% positions_term( +Data, -Fresh, -Arity ).
%
%  Create Fresh to be an uninstantiated 
%  version of Data. Arity is the *length* of Data.
%  It can be found from positions/2 independently,
%  but it is also provided here as it is often that 
%  we will be going through Fresh after calling this 
%  predicate.
%
%==
% positions_term( [1,2,3], Three ).
% positions_term( c(1,2,3), Three ).
% positions_term( x, X ).
%==
%
% @author nicos angelopoulos
% @version  0.1 2014/02/09
%
position_term( Data, Fresh ) :-
	position_term( Data, Fresh, _Arity ).

position_term( Data, Fresh, Arity ) :-
	position_type( Data, Type ),
	position_type_term( Type, Data, Fresh, Arity ).

position_type_term( list, Data, Fresh, Length ) :-
	length( Data, Length ),
	length( Fresh, Length ),
	!.
position_type_term( atomic, _Data, _Fresh, 1 ).
position_type_term( compound, Data, Fresh, Arity ) :-
	functor( Data, Name, Arity ),
	functor( Fresh, Name, Arity ).

%% positions_append( Data, Add, Total ).
%
%  Append 2 data items. Term wins list wins atomic. Left wins right.
%
% @author nicos angelopoulos
% @version 0.1 2014/02/09
% @version 0.2 2014/06/30
% @tbd string is incomplete
% @tbd use sub-calls to remove cuts.
% 
positions_append( Data, Add, Total ) :-
	position_type( Data, DType ),
	position_type( Add, AType ),
	positions_append( DType, AType, Data, Add, Total ).

positions_append( string, string, Data, Add, Total ) :-
	!,
	string_concat( Data, Add, Total ).
positions_append( list, list, Data, Add, Total ) :-
	!,
	append( Data, Add, Total ).
positions_append( list, compound, Data, Add, Total ) :-
	% temporary !
	!,
	Add =.. [Name|Args],
	append( Data, Args, TotArgs ),
	Total =.. [Name|TotArgs].
positions_append( list, atom, Data, Add, Total ) :-
	!,
	append( Data, [Add], Total ).
positions_append( compound, list, Data, Add, Total ) :-
	!,
	Data =.. [Name|Args],
	append( Args, Add, All ),
	Total =.. [Name|All].
positions_append( compound, compound, Data, Add, Total ) :-
	!,
	Data =.. [NameD|ArgsD],
	Add  =.. [_NameA|ArgsA],
	append( ArgsD, ArgsA, Args ),
	Total =.. [NameD|Args].
positions_append( compound, atom, Data, Add, Total ) :-
	!,
	Data =.. [Name|Args],
	append( Args, [Add], All ),
	append( Args, Add, All ),
	Total =.. [Name|All].
positions_append( atom, list, Data, Add, Total ) :-
	!,
	Total = [Data|Add].
positions_append( atom, compound, Data, Add, Total ) :-
	!,
	Add =.. [Name|Args],
	Total =.. [Name,Data|Args].
positions_append( atom, atom, Data, Add, Total ) :-
	atom_concat( Data, Add, Total ).

% for some operations numbers are best seen as atoms.
number_type_atom( number, DataIn, atom, Data ) :-
	!,
	atom_number( Data, DataIn ).
number_type_atom( Type, Data, Type, Data ).

number_type_converts( number, DataIn, Data ) :-
	!,
	atom_number( DataIn, Data ).
number_type_converts( _, Data, Data ).

type_term_elements_term( list, _Data, Elements, Elements ).
type_term_elements_term( atom, _Data, Elements, Atom ) :- 
	atom_codes( Atom, Elements ).
type_term_elements_term( compound, Data, Elements, Term ) :-
	functor( Data, Name, _Arity ),
	Term =.. [Name|Elements].
type_term_elements_term( string, _Data, Elements, String ) :-
	string_codes( String, Elements ).
